"""Web interface for LEX board game."""

# Non importiamo nulla qui per evitare il warning
# Se necessario, l'app può essere importata esplicitamente con:
# from lexweb.app import create_app

__all__ = []
